(function() {
var GridViewLayoutHelper = ASPx.CreateClass(ASPx.GridLayoutHelper, {
 constructor: function(grid) {
  this.constructor.prototype.constructor.call(this, grid);
  this.adaptiveLayoutInfo = null;
  this.mergeIntervals = null;
  this.groupLevels = null;
  this.spannedRowIds = [
   ASPx.GridViewConsts.GroupRowID,
   ASPx.GridViewConsts.GroupRowExpID,
   ASPx.GridViewConsts.PreviewRowID,
   ASPx.GridViewConsts.DetailRowID,
   ASPx.GridViewConsts.EmptyDataRowID,
   ASPx.GridViewConsts.EditingErrorItemID
  ];
  this.spannedTemplatedRowIds = [
   ASPx.GridViewConsts.DataRowID,
   ASPx.GridViewConsts.FooterRowID,
   ASPx.GridViewConsts.GroupFooterRowID,
   ASPx.GridViewConsts.FilterRowID,
   ASPx.GridViewConsts.EditingRowID
  ];
  this.spannedTemplatedCellIds = [
   ASPx.GridViewConsts.DataRowTemplateCellID,
   ASPx.GridViewConsts.FooterRowTemplateCellID,
   ASPx.GridViewConsts.GroupFooterRowTemplateCellID,
   ASPx.GridViewConsts.FilterRowTemplateCellID,
   ASPx.GridViewConsts.EditFormTableCellID
  ];
  this.cachedHasCellMerging = null;
  this.cachedHasServerGroupLevelInfo = null;
 },
 GetSpannedRowRegexp: function() {
  if(!this.spannedRowRegExp)
   this.spannedRowRegExp = new RegExp(this.GetSpannedElementPattern(this.spannedRowIds));
  return this.spannedRowRegExp;
 },
 GetSpannedTemplatedRowRegexp: function() {
  if(!this.spannedTemplatedRowRegExp)
   this.spannedTemplatedRowRegExp = new RegExp(this.GetSpannedElementPattern(this.spannedTemplatedRowIds));
  return this.spannedTemplatedRowRegExp;
 },
 GetSpannedTemplatedCellRegexp: function() {
  if(!this.spannedTemplatedCellRegExp)
   this.spannedTemplatedCellRegExp = new RegExp(this.GetSpannedElementPattern(this.spannedTemplatedCellIds));
  return this.spannedTemplatedCellRegExp;
 },
 GetSpannedElementPattern: function(idParts) {
  return "^" + this.grid.name + "_(" + idParts.join("|") + ")(-?\\d+)?(_\\d+)?$"; 
 },
 IsSpannedRow: function(row) { 
  if(!row) return false;
  var isTemplateRow = this.GetSpannedTemplatedRowRegexp().test(row.id);
  if(isTemplateRow) {
   var cell = this.GetSpannedCell(row);
   return cell && this.GetSpannedTemplatedCellRegexp().test(cell.id);
  }
  return this.GetSpannedRowRegexp().test(row.id) || this.IsBatchChangesViewCategoryRow(row);
 },
 IsBatchChangesViewCategoryRow: function(row) {
  return row.id.indexOf(ASPx.GridViewConsts.BatchChangesPreviewGroupRowID) > 0;
 },
 GetSpannedCell: function(row) { return this.grid.GetLastNonAdaptiveIndentCell(row); },
 IsArmatureRowRequired: function() { return this.grid.HasScrolling() || this.grid.AllowResizing(); }, 
 GetArmatureRows: function() { return this.GetAllTables().map(function(table) { return table.rows[0]; }); },
 GetColumnWidthInfo: function() { return this.grid.clientMoveColumnWidths; },
 GetColumnWidth: function(columnIndex) { 
  var info = this.GetColumnWidthInfo();
  if(!info) return null;
  var colWidth = info[columnIndex];
  return colWidth || "";
 },
 GetGridName: function() { return this.grid.name; },
 GetTotalIndentCount: function() { return this.grid.indentColumnCount; },
 GetGroupFooterVisibleIndices: function() { return this.grid.GetGroupFooterVisibleIndices(); },
 GetShowColumnHeaders: function() { return this.grid.showColumnHeaders; },
 GetShowFilterRow: function() { return !!this.grid.GetFilterRow(); },
 GetShowFooterRow: function() { return !!this.grid.GetFooterRow(); },
 GetShowEditingRow: function() { return this.IsInlineEditMode() && this.grid.GetEditingRow(); },
 IsBatchEditMode: function() { return this.grid.IsBatchEditMode(); },
 IsBatchChagesViewMode: function() { return this.grid.IsBatchChagesViewMode(); },
 IsInlineEditMode: function() { return this.grid.IsInlineEditMode(); },
 IsFixedTableLayout: function() {
  return this.grid.GetMainTable().style.tableLayout === "fixed";
 },
 HasMergedGroups: function() { return !!this.grid.hasMergedGroups; },
 GetAllTableRows: function() {
  var rows = [];
  var tables = this.GetAllTables();
  for(var i = 0; i < tables.length; i++)
   rows = rows.concat([].slice.call(tables[i].rows));
  return rows;
 },
 GetAllTables: function() {
  var grid = this.grid;
  var helper = grid.GetTableHelper();
  var headerTable = helper ? helper.GetHeaderTable() : null;
  var contentTable = grid.GetMainTable();
  var footerTable = helper ? helper.GetFooterTable() : null;
  return [ headerTable, contentTable, footerTable ].filter(function(table) { return !!table; });
 },
 GetMainElementCell: function() {
  return this.grid.GetMainElement().rows[0].cells[0];
 },
 Invalidate: function() { 
  ASPx.GridLayoutHelper.prototype.Invalidate.call(this);
  this.adaptiveLayoutInfo = null;
  this.mergeIntervals = null;
  this.groupLevels = null;
  this.cachedHasCellMerging = null;
  this.cachedHasServerGroupLevelInfo = null;
 },
 GetFixedColumnsHelper: function() { return this.grid.GetFixedColumnsHelper(); },
 GetIndentColumnCount: function() { return this.grid.indentColumnCount; },
 GetHasDetailRows: function() { return this.grid.hasDetailRows; },
 GetShowDetailButton: function() { return this.grid.showDetailButton; },
 GetHasDetailButton: function() { return this.GetHasDetailRows() && this.GetShowDetailButton(); },
 GetShowGroupButton: function() { return this.grid.showGroupButton; },
 GetHasGroupButton: function() { return this.GetGroupIndentsCount() > 0 && this.GetShowGroupButton() && !this.GetHasGroupRowTemplate(); },
 GetAdaptiveButtonPos: function() { return this.grid.adaptiveButtonPos; },
 GetHasFilterRowTemplate: function() { return this.grid.hasFilterRowTemplate; },
 GetHasDataRowTemplate: function() { return this.grid.hasDataRowTemplate; },
 GetHasGroupRowTemplate: function() { return this.grid.hasGroupRowTemplate; },
 GetHasGroupFooterRowTemplate: function() { return this.grid.hasGroupFooterRowTemplate; },
 GetHasFooterRowTemplate: function() { return this.grid.hasFooterRowTemplate; },
 IsDataRow: function(row) {
  if(!row || !row.id) return false;
  var dataRowIdPrefix = this.grid.name + "_" + ASPx.GridViewConsts.DataRowID;
  return row.id.indexOf(dataRowIdPrefix) > -1;
 },
 IsAdaptiveCell: function(cell) {
  var adaptiveClasses = [ASPx.GridViewConsts.AdaptiveDetailDataCellCssClass, ASPx.GridViewConsts.AdaptiveDetailCommandCellCssClass];
  return adaptiveClasses.some(function(className) { return ASPx.ElementHasCssClass(cell, className); });
 }, 
 IsServiceCell: function(cell) {
  var serviceCssClasses = [ASPx.GridViewConsts.IndentCellCssClass, ASPx.GridViewConsts.HorizontalExtraCellClassName, ASPx.GridViewConsts.AdaptiveIndentCellCssClass];
  return serviceCssClasses.some(function(className) { return ASPx.ElementHasCssClass(cell, className); });
 },
 HasServerGroupLevelInfo: function() {
  if(this.cachedHasServerGroupLevelInfo === null) {
   var groupLevelInfo = this.GetGroupLevelServerInfo();
   this.cachedHasServerGroupLevelInfo = groupLevelInfo && Object.keys(groupLevelInfo).length > 0;
  }
  return this.cachedHasServerGroupLevelInfo;
 },
 HasCellMerging: function() { 
  if(this.cachedHasCellMerging === null) {
   var mergedCellInfo = this.GetMergedCellServerInfo();
   this.cachedHasCellMerging = mergedCellInfo && Object.keys(mergedCellInfo).length > 0;
  }
  return this.cachedHasCellMerging;
 },
 GetGroupLevelServerInfo: function() { return this.grid.stateObject.groupLevelState || { }; },
 SetGroupLevelServerInfo: function(info) { this.grid.stateObject.groupLevelState = info; },
 GetGroupRowLevel: function(visibleIndex) { 
  this.EnsureGroupLevelInfo();
  if(!this.groupLevels) return -1;
  var level = this.groupLevels[visibleIndex];
  return ASPx.IsExists(level) ? level : -1;
 },
 EnsureGroupLevelInfo: function() {
  if(this.groupLevels || !this.HasServerGroupLevelInfo()) 
   return;
  var serverInfo = this.GetGroupLevelServerInfo();
  this.groupLevels = { };
  for(var level in serverInfo) {
   if(!serverInfo.hasOwnProperty(level)) continue;
   level = parseInt(level);
   var intervals = serverInfo[level];
   for(var i = 0; i < intervals.length; i++) {
    var interval = intervals[i];
    var start = interval[0];
    var end = interval[1];
    for(var visibleIndex = start; visibleIndex <= end; visibleIndex++)
     this.groupLevels[visibleIndex] = level;
   }
  }
 },
 GetGroupRowColSpan: function(visibleIndex) { 
  var groupLevel = this.GetGroupRowLevel(visibleIndex);
  if(!ASPx.IsExists(groupLevel))
   return -1;
  var groupIndentsCount = this.GetGroupIndentsCount();
  var layoutWidth = this.GetLayoutWidth();
  var detailButtonWidth = this.GetHasDetailButton() ? 1 : 0;
  var groupButtonWidth = this.GetHasGroupButton() ? 1 : 0;
  return layoutWidth + detailButtonWidth + groupIndentsCount - groupLevel - groupButtonWidth;
 },
 GetGroupFooterRowColSpan: function(visibleIndex) { 
  var layoutWidth = this.GetLayoutWidth();
  var groupLevel = this.GetGroupRowLevel(visibleIndex);
  var footerIndentCount = this.GetGroupIndentsCount() - groupLevel - 1;
  return layoutWidth + footerIndentCount;
 },
 GetGroupIndentsCount: function() {
  return this.HasMergedGroups() ? 1 : this.GetColumnHelper().GetGroupColumnCount();
 },
 GetIndentCellCount: function(visibleIndex, levelIndex, rowType) { 
  var result = this.GetIndentCellCountCore(visibleIndex, rowType);
  if(levelIndex === 0)
   return Math.max(0, result);
  var groupIndentsCount = this.GetGroupIndentsCount();
  var detailButtonWidth = this.GetHasDetailButton() ? 1 : 0;
  if(rowType === ASPx.GridViewRowType.Footer && groupIndentsCount > 0) 
   result--;
  switch(rowType) {
   case ASPx.GridViewRowType.Footer:
   case ASPx.GridViewRowType.Filter:
   case ASPx.GridViewRowType.Data:
   case ASPx.GridViewRowType.BatchEtalon:
   case ASPx.GridViewRowType.GroupFooter:
    if(detailButtonWidth > 0)
     result--;
    break;
   case ASPx.GridViewRowType.Header:
    result = 0;
  }
  return Math.max(0, result);
 },
 GetIndentCellCountCore: function(visibleIndex, rowType) {
  var groupButtonWidth = this.GetHasGroupButton() ? 1 : 0;
  var detailButtonWidth = this.GetHasDetailButton() ? 1 : 0;
  var leftAdaptiveButtonWidth = this.GetAdaptiveButtonPos() === ASPx.GridViewAdaptiveDetailButtonPosition.Left ? 1 : 0;
  var groupIndentsCount = this.GetGroupIndentsCount();
  var groupLevel = this.GetGroupRowLevel(visibleIndex);
  var totalIndentCount = groupIndentsCount + leftAdaptiveButtonWidth + detailButtonWidth;
  switch(rowType) {
   case ASPx.GridViewRowType.Header:
   case ASPx.GridViewRowType.Filter:
   case ASPx.GridViewRowType.Data:
   case ASPx.GridViewRowType.BatchEtalon:
   case ASPx.GridViewRowType.InlineEdit:
    return totalIndentCount;
   case ASPx.GridViewRowType.Footer:
    return leftAdaptiveButtonWidth + detailButtonWidth + (groupIndentsCount > 0 ? 1 : 0);
   case ASPx.GridViewRowType.Group:
    return groupLevel + groupButtonWidth;
   case ASPx.GridViewRowType.GroupFooter:
    var footerIndentCount = groupIndentsCount - groupLevel - 1;
    var dataIndentCount = totalIndentCount - footerIndentCount;
    var footerIndentWidth = (footerIndentCount > 0 && !this.GetHasGroupFooterRowTemplate()) ? 1 : 0;
    return dataIndentCount + footerIndentWidth;
  }
  return 0;
 },
 GetRowInfo: function(rowID) {
  if(!rowID) return null;
  var pattern = "^" + this.GetGridName() + "_([a-z]+)(-?\\d+)?(_\\d+)?$";
  var regex = new RegExp(pattern, "i");
  var matches = regex.exec(rowID);
  if(!matches || matches.length !== 4)
   return null;
  var firstNumber = parseInt(matches[2]);
  var secondNumber = matches[3] !== undefined ? parseInt(matches[3].substr(1)) : NaN;
  var rowType = this.GetRowTypeByPartialID(matches[1]);
  if(rowType === null)
   return null;
  var visibleIndex = -1;
  var level = 0;
  switch(rowType) {
   case ASPx.GridViewRowType.Header:
    if(!isNaN(secondNumber)) return null;
    level = firstNumber;
    break;
   case ASPx.GridViewRowType.Filter:
   case ASPx.GridViewRowType.Footer:
   case ASPx.GridViewRowType.InlineEdit:
   case ASPx.GridViewRowType.BatchEtalon:
    if(!isNaN(firstNumber)) return null;
    level = secondNumber;
    break;
   case ASPx.GridViewRowType.Group:
    if(isNaN(firstNumber) || !isNaN(secondNumber)) return null;
    visibleIndex = firstNumber;
    break;
   case ASPx.GridViewRowType.Data:
   case ASPx.GridViewRowType.GroupFooter:
    if(isNaN(firstNumber)) return null;
    visibleIndex = firstNumber;
    level = secondNumber;
    break;
  }
  if(isNaN(level))
   level = 0;
  return { rowType: rowType, visibleIndex: visibleIndex, level: level };
 },
 GetRowTypeByPartialID: function(partialID) {
  switch(partialID) {
   case ASPx.GridViewConsts.HeaderRowID:
    return ASPx.GridViewRowType.Header;
   case ASPx.GridViewConsts.FilterRowID:
    return ASPx.GridViewRowType.Filter;
   case ASPx.GridViewConsts.DataRowID:
    return ASPx.GridViewRowType.Data;
   case ASPx.GridViewConsts.GroupRowID:
   case ASPx.GridViewConsts.GroupRowExpID:
    return ASPx.GridViewRowType.Group;
   case ASPx.GridViewConsts.GroupFooterRowID:
    return ASPx.GridViewRowType.GroupFooter;
   case ASPx.GridViewConsts.FooterRowID:
    return ASPx.GridViewRowType.Footer;
   case ASPx.GridViewConsts.EditingRowID:
    return ASPx.GridViewRowType.InlineEdit; 
   case ASPx.GridViewConsts.DataRowID + this.GetEtalonBatchRecordPostfix():
    return ASPx.GridViewRowType.BatchEtalon;
  }
  return null;
 },
 GetEtalonBatchRecordPostfix: function() { return this.grid.EtalonBatchRecordPostFix; },
 GetRowCellMap: function(info) {
  if(!info || info.rowType === ASPx.GridViewRowType.Group) 
   return null;
  var layout = null;
  if(info.rowType === ASPx.GridViewRowType.Header)
   layout = this.GetHeaderLayout();
  else 
   layout = info.rowType === ASPx.GridViewRowType.Data ? this.GetDataRowLayout(info.visibleIndex) : this.GetEtalonDataRowLayout();
  var result = { };
  var indentCount = this.GetIndentCellCount(info.visibleIndex, info.level, info.rowType);
  var level = layout[info.level];
  for(var i = 0; i < level.length; i++)
   result[level[i].columnIndex] = indentCount + i;
  return result;
 },
 CopyIndentCells: function(row, cloneRow, visibleIndex, cloneLevelIndex, rowType) {
  if(!this.IsSpannedRow(row)) {
   this.CopyIndentCellsFromLeft(row, cloneRow, visibleIndex, cloneLevelIndex, rowType);
   this.CopyIndentCellsFromRight(row, cloneRow, visibleIndex, cloneLevelIndex, rowType);
  }
 },
 CopyIndentCellsFromLeft: function(row, cloneRow, visibleIndex, cloneLevelIndex, rowType) {
  if(rowType === ASPx.GridViewRowType.Header)
   return;
  var firstRowIndentCellCount = this.GetIndentCellCount(visibleIndex, 0, rowType);
  var cloneRowIndentCellCount = this.GetIndentCellCount(visibleIndex, cloneLevelIndex, rowType);
  var indentCellIndices = ASPx.Data.Range(firstRowIndentCellCount, 0);
  var detailIndentCellIndex = this.GetDetailIndentCellIndex(visibleIndex, 0, rowType);
  var footerGroupIndentCellIndex = this.GetFooterGroupIndentCellIndex(visibleIndex, 0, rowType);
  ASPx.Data.ArrayRemove(indentCellIndices, detailIndentCellIndex);
  ASPx.Data.ArrayRemove(indentCellIndices, footerGroupIndentCellIndex);
  if(indentCellIndices.length !== cloneRowIndentCellCount)
   throw "Wrong indent cell count for clonned row";
  this.CloneIndentCells(row, cloneRow, indentCellIndices);
 },
 CopyIndentCellsFromRight: function(row, cloneRow, visibleIndex, cloneLevelIndex, rowType) { 
  var layoutWidth = this.GetLayoutWidth(rowType === ASPx.GridViewRowType.Header);
  var indentCellCount = this.GetIndentCellCount(visibleIndex, 0, rowType);
  var rightIndentCellIndex = indentCellCount + layoutWidth;
  var indentCellIndices = ASPx.Data.Range(row.cells.length - rightIndentCellIndex, rightIndentCellIndex);
  this.CloneIndentCells(row, cloneRow, indentCellIndices);
 },
 CloneIndentCells: function(row, cloneRow, cellIndices) {
  var targetElement = cloneRow.insertCell();
  for(var i = 0; i < cellIndices.length; i++) {
   var cellIndex = cellIndices[i];
   var cloneCell = row.cells[cellIndex].cloneNode();
   cloneRow.insertBefore(cloneCell, targetElement);
  }
  ASPx.RemoveElement(targetElement);
 },
 GetDetailIndentCellIndex: function(visibleIndex, levelIndex, rowType) {
  if(!this.GetHasDetailButton() || levelIndex > 0) 
   return -1;
  var indentCellCount = this.GetIndentCellCount(visibleIndex, 0, rowType);
  return indentCellCount - 1;
 },
 GetFooterGroupIndentCellIndex: function(visibleIndex, levelIndex, rowType) {
  if(rowType !== ASPx.GridViewRowType.Footer || this.GetHasFooterRowTemplate() || levelIndex > 0) 
   return -1;
  var indentCellCount = this.GetIndentCellCount(visibleIndex, 0, rowType);
  var result = indentCellCount - 1;
  if(this.GetHasDetailButton())
   result--;
  return result;
 },
 UpdateIndentCellRowSpan: function(visibleIndex, rowType) {
  var indentCellCount = this.GetIndentCellCount(visibleIndex, 0, rowType);
  var totalIndentCellIndices = ASPx.Data.Range(indentCellCount, 0);
  var detailIndentCellIndex = this.GetDetailIndentCellIndex(visibleIndex, 0, rowType);
  var footerGroupIndentCellIndex = this.GetFooterGroupIndentCellIndex(visibleIndex, 0, rowType);
  var cellIndices = rowType === ASPx.GridViewRowType.Header ? totalIndentCellIndices : [ detailIndentCellIndex, footerGroupIndentCellIndex ];
  var height = this.GetLayoutHeight(rowType === ASPx.GridViewRowType.Header);
  var row = this.GetRow(visibleIndex, 0, rowType);
  for(var i = 0; i < cellIndices.length; i++) {
   var cellIndex = cellIndices[i];
   if(cellIndex >= 0 && cellIndex < row.cells.length)
    row.cells[cellIndex].rowSpan = height;
  }
 },
 GetColumnHelper: function() { return this.grid.GetColumnHelper(); },
 GetAdaptivityHelper: function() { return this.grid.GetAdaptivityHelper(); },
 GetMergedCellServerInfo: function() { return this.grid.stateObject.mergedCellState || { }; },
 SetMergedCellServerInfo: function(info) { this.grid.stateObject.mergedCellState = info; },
 GetHeaderLayout: function() { return this.GetColumnHelper().GetHeaderLayout(); },
 GetEtalonDataRowLayout: function() { return this.GetColumnHelper().GetDataRowLayout(); },
 GetDataRowLayout: function(visibleIndex) {
  var layout = this.GetEtalonDataRowLayout();
  if(!ASPx.IsExists(visibleIndex) || !this.HasCellMerging())
   return layout;
  var newLayout = [ ];
  for(var i = 0; i < layout.length; i++) {
   var level = layout[i].slice();
   for(var j = level.length - 1; j >= 0; j--) {
    if(this.IsHiddenMergedCell(visibleIndex, level[j].columnIndex))
     level.splice(j, 1); 
   }
   newLayout.push(level);
  }
  return newLayout;
 },
 GetHasRowTemplate: function(rowType) {
  switch(rowType) {
   case ASPx.GridViewRowType.Filter:
    return this.GetHasFilterRowTemplate();
   case ASPx.GridViewRowType.Data:
    return this.GetHasDataRowTemplate();
   case ASPx.GridViewRowType.GroupFooter:
    return this.GetHasGroupFooterRowTemplate();
   case ASPx.GridViewRowType.Footer:
    return this.GetHasFooterRowTemplate();
  }
  return false;
 },
 GetRow: function(visibleIndex, levelIndex, rowType) {
  switch(rowType) {
   case ASPx.GridViewRowType.Header:
    return this.grid.GetHeaderRow(levelIndex);
   case ASPx.GridViewRowType.Filter:
    return this.grid.GetFilterRow(levelIndex);
   case ASPx.GridViewRowType.Data:
    return this.grid.GetDataRow(visibleIndex, levelIndex);
   case ASPx.GridViewRowType.GroupFooter:
    return this.grid.GetGroupFooterRow(visibleIndex, levelIndex);
   case ASPx.GridViewRowType.Footer:
    return this.grid.GetFooterRow(levelIndex);
   case ASPx.GridViewRowType.InlineEdit:
    return this.grid.GetEditingRow(this.grid, levelIndex);
   case ASPx.GridViewRowType.BatchEtalon:
    return this.grid.GetBatchEtalonRow(levelIndex);   
  }
  return null;
 },
 GetStretchedRowColsSpan: function(visibleIndex, rowType) { 
  var layoutWidth = this.GetLayoutWidth();
  switch(rowType) {
   case ASPx.GridViewRowType.Filter:
   case ASPx.GridViewRowType.Data:
   case ASPx.GridViewRowType.Footer:
    return layoutWidth;
   case ASPx.GridViewRowType.Group:
    return this.GetGroupRowColSpan(visibleIndex);
   case ASPx.GridViewRowType.GroupFooter:
    return this.GetGroupFooterRowColSpan(visibleIndex);
  }
  return 0;
 },
 GetCell: function(visibleIndex, columnIndex, rowType) {
  switch(rowType) {
   case ASPx.GridViewRowType.Header:
    return this.GetHeaderCell(columnIndex);
   case ASPx.GridViewRowType.Filter:
    return this.GetFilterCell(columnIndex);
   case ASPx.GridViewRowType.Data:
    return this.GetDataCell(visibleIndex, columnIndex);
   case ASPx.GridViewRowType.GroupFooter:
    return this.GetGroupFooterCell(visibleIndex, columnIndex);
   case ASPx.GridViewRowType.Footer:
    return this.GetFooterCell(columnIndex);
   case ASPx.GridViewRowType.BatchEtalon:
    return this.GetBatchEtalonCell(visibleIndex, columnIndex);
   case ASPx.GridViewRowType.InlineEdit:
    return this.GetInlineEditCell(columnIndex);
  }
  return null;
 },
 UpdateStretchedRowsColSpan: function() {
  var helper = this.GetFixedColumnsHelper();
  this.UpdateStretchedRowsColSpanCore(helper ? helper.hiddenColumnCount : 0);
 },
 UpdateStretchedRowsColSpanCore: function(columnsToHide) {
  var groupRowRegexp = new RegExp(this.GetSpannedElementPattern([ ASPx.GridViewConsts.GroupRowID, ASPx.GridViewConsts.GroupRowExpID ]));
  var groupFooterRegexp = new RegExp(this.GetSpannedElementPattern([ ASPx.GridViewConsts.GroupFooterRowID ]));
  var totalSpan = this.GetLayoutWidth(false);
  var stretchedRows = this.GetAllTableRows().filter(function(row) { return this.IsSpannedRow(row); }.aspxBind(this));
  for(var i = 0; i < stretchedRows.length; i++) {
   var row = stretchedRows[i];
   var cell = this.GetSpannedCell(row);
   var colSpan = totalSpan;
   if(groupRowRegexp.test(row.id)) {
    var matches = row.id.match(groupRowRegexp);
    var visibleIndex = parseInt(matches[2]);
    colSpan = this.GetStretchedRowColsSpan(visibleIndex, ASPx.GridViewRowType.Group);
   }
   if(groupFooterRegexp.test(row.id)) {
    var matches = row.id.match(groupFooterRegexp);
    var visibleIndex = parseInt(matches[2]);
    colSpan = this.GetStretchedRowColsSpan(visibleIndex, ASPx.GridViewRowType.GroupFooter);
   }
   ASPx.Attr.ChangeCellSpanCount(cell, colSpan - columnsToHide, true);
  }
 },
 GetHeaderCell: function(columnIndex) { return this.grid.GetHeader(columnIndex); },
 GetFilterCell: function(columnIndex) { return this.GetCellCore(-1, columnIndex, ASPx.GridViewRowType.Filter); },
 GetDataCell: function(visibleIndex, columnIndex) { return this.GetCellCore(visibleIndex, columnIndex, ASPx.GridViewRowType.Data); },
 GetGroupFooterCell: function(visibleIndex, columnIndex) { return this.GetCellCore(visibleIndex, columnIndex, ASPx.GridViewRowType.GroupFooter); },
 GetFooterCell: function(columnIndex) { return this.GetCellCore(-1, columnIndex, ASPx.GridViewRowType.Footer); },
 GetBatchEtalonCell: function(visibleIndex, columnIndex) { return this.GetCellCore(visibleIndex, columnIndex, ASPx.GridViewRowType.BatchEtalon); },
 GetInlineEditCell: function(columnIndex) { 
  if(this.IsInlineEditMode())
   return this.GetCellCore(-1, columnIndex, ASPx.GridViewRowType.InlineEdit); 
  return null;
 },
 GetCellCore: function(visibleIndex, columnIndex, rowType) {
  if(this.GetHasRowTemplate(rowType)) return null;
  columnIndex = parseInt(columnIndex);
  var layout = this.GetEtalonDataRowLayout();
  var pos = this.GetColumnHelper().FindColumnLayoutPos(layout, columnIndex);
  if(pos.levelIndex < 0 || pos.cellIndex < 0)
   return null;
  visibleIndex = this.ResolveMergedCellVisibleIndex(visibleIndex, columnIndex);
  var cellIndex = pos.cellIndex + this.GetIndentCellCount(visibleIndex, pos.levelIndex, rowType);
  if(this.HasCellMerging() && (rowType === ASPx.GridViewRowType.Data || rowType === ASPx.GridViewRowType.BatchEtalon))
   cellIndex -= this.CalcHiddenMergedCellCount(visibleIndex, columnIndex);
  var row = this.GetRow(visibleIndex, pos.levelIndex, rowType);
  return row && cellIndex >= 0 && cellIndex < row.cells.length ? row.cells[cellIndex] : null;
 },
 GetHeaderRows: function() {
  var height = this.GetShowColumnHeaders() ? this.GetLayoutHeight(true) : 0;
  return ASPx.Data.Range(height, 0).map(function(i) { return this.grid.GetHeaderRow(i); }.bind(this));
 },
 GetFilterRows: function() {
  var height = this.GetLayoutHeight(false);
  return ASPx.Data.Range(height, 0).map(function(i) { return this.grid.GetFilterRow(i); }.bind(this));
 },
 GetLayoutWidth: function(isHeader) {
  var layout = isHeader ? this.GetHeaderLayout() : this.GetEtalonDataRowLayout();
  return this.GetLayoutWidthCore(layout);
 },
 GetLayoutWidthCore: function(layout) {
  if(!layout || layout.length === 0) 
   return 0;
  var firstLevel = layout[0];
  if(!firstLevel || firstLevel.length === 0)
   return 0;
  return firstLevel.reduce(function(prevValue, node) { return prevValue + node.colSpan; }, 0);
 },
 GetLayoutHeight: function(isHeader) {
  var layout = isHeader ? this.GetHeaderLayout() : this.GetEtalonDataRowLayout();
  return this.GetLayoutHeightCore(layout);
 },
 GetLayoutHeightCore: function(layout) {
  return layout ? layout.length : 0;
 },
 GetColumnIndex: function(cellIndex, visibleIndex, levelIndex) { 
  if(!levelIndex || levelIndex < 0)
   levelIndex = 0;
  var layout = this.GetDataRowLayout(visibleIndex); 
  if(layout.length <= levelIndex)
   return -1;
  var level = layout[levelIndex];
  cellIndex -= this.GetIndentCellCount(visibleIndex, levelIndex, ASPx.GridViewRowType.Data);
  return cellIndex >= 0 && cellIndex < level.length ? level[cellIndex].columnIndex : -1;
 },
 GetDataCellIndex: function(columnIndex, visibleIndex) { 
  var layout = this.GetDataRowLayout(visibleIndex);
  var pos = this.GetColumnHelper().FindColumnLayoutPos(layout, columnIndex);
  if(pos.levelIndex < 0 || pos.cellIndex < 0)
   return -1;
  return pos.cellIndex + this.GetIndentCellCount(visibleIndex, pos.levelIndex, ASPx.GridViewRowType.Data);
 },
 IsMergedCell: function(visibleIndex, columnIndex) { 
  return !!this.GetMergeInterval(visibleIndex, columnIndex);
 },
 IsHiddenMergedCell: function(visibleIndex, columnIndex) {
  return visibleIndex !== this.ResolveMergedCellVisibleIndex(visibleIndex, columnIndex);
 },
 IsMergedCellLastRow: function(visibleIndex, columnIndex) { 
  var interval = this.GetMergeInterval(visibleIndex, columnIndex);
  return !!interval && interval[1] === visibleIndex;
 },
 ResolveMergedCellVisibleIndex: function(visibleIndex, columnIndex) {
  if(!this.HasCellMerging()) return visibleIndex;
  var interval = this.GetMergeInterval(visibleIndex, columnIndex);
  return interval ? interval[0] : visibleIndex;
 },
 CalcHiddenMergedCellCount: function(visibleIndex, columnIndex) {
  var layout = this.GetEtalonDataRowLayout();
  var pos = this.GetColumnHelper().FindColumnLayoutPos(layout, columnIndex);
  if(pos.levelIndex < 0 || pos.cellIndex < 0)
   return 0;
  var result = 0;
  var level = layout[pos.levelIndex];
  for(var j = 0; j < pos.cellIndex; j++) {
   if(this.IsHiddenMergedCell(visibleIndex, level[j].columnIndex))
    result++;
  }
  return result;
 },
 GetMergeInterval: function(visibleIndex, columnIndex) {
  this.EnsureMergeIntervals();
  var columnIntervals = this.mergeIntervals[columnIndex];
  return columnIntervals ? columnIntervals[visibleIndex] : null;
 },
 EnsureMergeIntervals: function() {
  if(this.mergeIntervals) return;
  this.mergeIntervals = { };
  var serverInfo = this.GetMergedCellServerInfo();
  for(var columnIndex in serverInfo) {
   if(!serverInfo.hasOwnProperty(columnIndex)) continue;
   var columnIntervals = { };
   var serverIntervals = serverInfo[columnIndex];
   for(var i = 0; i < serverIntervals.length; i++) {
    var interval = serverIntervals[i];
    var start = interval[0];
    var end = interval[1];
    for(var j = start; j <= end; j++)
     columnIntervals[j] = interval;
   }
   this.mergeIntervals[parseInt(columnIndex)] = columnIntervals;
  }
 },
 GetAdaptiveRowLayoutInfo: function(visibleIndex) {
  if(!this.adaptiveLayoutInfo) 
   this.adaptiveLayoutInfo = this.CreateAdaptiveLayoutInfo();
  var baseAdaptiveLayout = this.adaptiveLayoutInfo.adaptiveRowLayout || [ ];
  return { 
   dataRowLayout: this.adaptiveLayoutInfo.dataRowLayout || this.GetEtalonDataRowLayout(),
   adaptiveRowLayout : this.IsAdaptiveDetailExpanded(visibleIndex) ? baseAdaptiveLayout : [ ]
  };
 },
 CreateAdaptiveLayoutInfo: function() {
  var adaptiveHelper = this.GetAdaptivityHelper();
  if(!adaptiveHelper) 
   return { };
  var dataRowLayout = [ ];
  var adaptiveColumnIndices = [ ];
  var columnHelper = this.GetColumnHelper();
  var isFormLayoutMode = adaptiveHelper.IsAdaptiveFormLayoutMode();
  if(adaptiveHelper.IsAdaptivityMode()) {
   var hasAnyAdaptive = adaptiveHelper.HasAnyAdaptiveElement();
   if(hasAnyAdaptive) {
    dataRowLayout = [ ];
    adaptiveColumnIndices = isFormLayoutMode ? adaptiveHelper.GetAdaptiveLayoutColumnIndices() : columnHelper.GetVisibleColumnIndices();
   } else {
    dataRowLayout = this.GetEtalonDataRowLayout();
    adaptiveColumnIndices = [ ];
   }
  } else {
   var adaptiveInfo = columnHelper.GetAdaptiveRowInfo();
   dataRowLayout = adaptiveInfo.dataRowLayout;
   var columnIndicesForAdaptiveIndicesOrder = isFormLayoutMode ? adaptiveHelper.GetAdaptiveLayoutColumnIndices() : columnHelper.GetVisibleColumnIndices();
   for(var i = 0; i < columnIndicesForAdaptiveIndicesOrder.length; i++) {
    var columnIndex = columnIndicesForAdaptiveIndicesOrder[i];
    if(adaptiveInfo.adaptiveColumnIndices.indexOf(columnIndex) > -1)
     adaptiveColumnIndices.push(columnIndex);
   }
  }
  return { dataRowLayout: dataRowLayout, adaptiveRowLayout: this.CreateAdaptiveRowLayout(adaptiveColumnIndices) };
 },
 CreateAdaptiveRowLayout: function(columnIndices) {
  var layout = [ ];
  var width = this.GetAdaptiveLayoutWidth();
  var height = Math.ceil(columnIndices.length / width);
  for(var i = 0; i < height; i++) {
   var level = [ ];
   for(var j = 0; j < width; j++) {
    var index = i * width + j;
    if(index < columnIndices.length)
     level.push({ columnIndex: columnIndices[index], colSpan: 1, rowSpan: 1 });
    else
     level[level.length - 1].colSpan++;
   }
   layout.push(level);
  }
  return layout;
 },
 IsAdaptiveDetailExpanded: function(visibleIndex) {
  var helper = this.GetAdaptivityHelper();
  if(!helper || !helper.HasAnyAdaptiveElement())
   return false;
  return helper.IsAdaptivityMode() || helper.IsAdaptiveDetailExpanded(visibleIndex);
 },
 GetAdaptiveLayoutWidth: function() {
  var helper = this.GetAdaptivityHelper();
  return helper.GetAdaptiveLayoutColumnCount();
 },
 AreDataRowsHidden: function() {
  var helper = this.GetAdaptivityHelper();
  return helper && helper.IsAdaptivityMode() && helper.HasAnyAdaptiveElement();
 },
 GetAdaptiveDataCellByChild: function(dataItem, element) {
  while(element && element.tagName !== "BODY") {
   if(ASPx.ElementHasCssClass(element, ASPx.GridViewConsts.AdaptiveDetailDataCellCssClass))
    return element;
   element = element.parentNode;
  }
  return null;
 },
 IsAdaptiveDetailRow: function(row) {
  if(!row || !row.id) return false;
  var grid = this.grid;
  var indexInfo = grid.tryGetNumberFromEndOfString(row.id);
  if(!indexInfo.success) return false;
  return row.id === this.GetAdaptiveDetailRowID(indexInfo.value);
 },
 GetAdaptiveDetailRowID: function(index) { return this.grid.name + "_" + ASPx.GridViewConsts.AdaptiveDetailRowID + index.toString(); },
 UpdateOnErrorRowInserting: function(visibleIndex, errorRow) {
  var mergedIndices = this.GetErrorRowMergedColumnIndices(visibleIndex);
  if(!mergedIndices.length) return;
  this.CreateErrorRowExtraCells(errorRow, this.GetErrorRowColSpans(mergedIndices));
  this.UpdateErrorRowIntersectedMergedCells(visibleIndex, 1);
  errorRow.dxMergedCellsIntersect = true;
 },
 UpdateOnErrorRowDeleting: function(visibleIndex, errorRow) {
  if(errorRow && errorRow.dxMergedCellsIntersect)
   this.UpdateErrorRowIntersectedMergedCells(visibleIndex, -1);
 },
 GetErrorRowMergedColumnIndices: function(visibleIndex) {
  return this.GetColumnHelper().GetVisibleColumnIndices().filter(function(columnIndex) { 
   return this.IsMergedCell(visibleIndex, columnIndex) && !this.IsMergedCellLastRow(visibleIndex, columnIndex); 
  }.bind(this));
 },
 CreateErrorRowExtraCells: function(errorRow, colSpans) {
  var errorRowCells = Array(colSpans.length);
  errorRowCells[0] = this.grid.GetLastNonAdaptiveIndentCell(errorRow);
  for(var i = 0; i < colSpans.length; i++) {
   if(!errorRowCells[i]) {
    var td = ASPx.CreateHtmlElement("TD", { className: ASPx.GridViewConsts.GridViewMarkerCssClass });
    ASPx.InsertElementAfter(td, errorRowCells[i - 1]);
    errorRowCells[i] = td;
   }
   ASPx.Attr.ChangeCellSpanCount(errorRowCells[i], colSpans[i], true);
  }
 },
 GetErrorRowColSpans: function(mergedIndices) {
  var columnIndices = this.GetColumnHelper().GetVisibleColumnIndices();
  var colSpans = [ ];
  for(var i = 0; i < columnIndices.length; i++) { 
   var isMergedCell = mergedIndices.indexOf(columnIndices[i]) > -1;
   if(isMergedCell || colSpans.length === 0)
    colSpans.push(0);
   if(!isMergedCell)
    ++colSpans[colSpans.length - 1];
  }
  return colSpans.filter(function(span) { return span > 0; });
 },
 UpdateErrorRowIntersectedMergedCells: function(visibleIndex, inc) {
  var indices = this.GetErrorRowMergedColumnIndices(visibleIndex);
  for(var i = 0; i <indices.length; i++) {
   var mergedCell = this.GetDataCell(visibleIndex, indices[i]);
   ASPx.Attr.ChangeCellSpanCount(mergedCell, mergedCell.rowSpan + inc);
  }
 },
 GetDataCellByChild: function(element) {
  return ASPx.GetParent(element, function(item) {
   if(this.IsServiceCell(item))
    return false;
   return this.IsAdaptiveCell(item) || this.IsDataRow(item.parentNode);
  }.aspxBind(this)); 
 },
 GetBatchChangesRecordVisibleIndices: function() { return this.grid.GetBatchChangesRecordVisibleIndices(); },
 GetGridElementsToChangeVisibilityInBatchChangesView: function() {
  var grid = this.grid;   
  var mainElement = grid.GetMainElement();
  var elements = [ ];
  elements.push(ASPx.GetNodesByPartialClassName(mainElement, "dxgvPagerBottomPanel")[0]);
  elements.push(ASPx.GetNodesByPartialClassName(mainElement, "dxgvPagerTopPanel")[0]);
  elements.push(grid.GetGroupPanel());
  elements.push(grid.GetChildElement("DXSearchPanel"));
  elements.push(grid.GetChildElement("DXFilterBar"));
  elements.push(grid.GetAdaptiveGroupPanel());
  elements.push(grid.GetAdaptiveHeaderPanel());
  elements.push(grid.GetAdaptiveFooterPanel());
  elements.push(grid.GetAdaptiveFooterPanel());
  elements = elements.concat(this.GetRowsToChangeVisibilityInBatchChangesView());
  elements = elements.concat(this.GetIndentCellsToChangeVisibilityInBatchChangesView());
  var commandColumnIndices = this.GetColumnHelper().GetVisibleColumnIndices().filter(function(columnIndex) { return grid.GetColumn(columnIndex).isCommandColumn; });
  var selectAllElements = commandColumnIndices.map(function(columnIndex) {
   var el = grid.GetSelectAllBtn(columnIndex);
   return el && el.parentNode.parentNode; 
  });
  elements = elements.concat(selectAllElements);
  var filterHelper = grid.GetFilterHelper();
  if(filterHelper)
   elements = elements.concat(filterHelper.GetHeaderFilterButtons().map(function(b) { return b && b.parentNode; })); 
  return elements.filter(function(el) { return !!el; });
 },
 GetIndentCellsToChangeVisibilityInBatchChangesView: function () {
  var grid = this.grid;
  var mainElement = grid.GetMainElement();
  var cells = [ ];
  var detailIndents =  ASPx.GetNodesByPartialClassName(mainElement, "dxgvDI");   
  cells = cells.concat(detailIndents);
  var groupIndentCells = ASPx.GetNodesByPartialClassName(mainElement, "dxgvGI");
  cells = cells.concat(groupIndentCells);
  var headerRow = grid.GetHeaderRow(0);
  if(headerRow) {
   var groupIndentsCount = this.GetGroupIndentsCount();
   var detailButtonWidth = this.GetHasDetailButton() ? 1 : 0;
   var indentCount = groupIndentsCount + detailButtonWidth;
   for(var i = 0; i < indentCount; i++)
    cells.push(headerRow.cells[i]);
   }
  var dataRowIndents = ASPx.GetNodesByPartialClassName(mainElement, ASPx.GridViewConsts.DetailButtonCellCssClass);
  dataRowIndents = dataRowIndents.concat(ASPx.GetNodesByPartialClassName(mainElement, ASPx.GridViewConsts.IndentCellCssClass));
  cells = cells.concat(dataRowIndents);
  return cells;
 },
 GetRowsToChangeVisibilityInBatchChangesView: function() {
  var rows = [ ];
  var grid = this.grid;
  rows = rows.concat(this.GetFilterRows());
  var indices = grid.GetCurrentPageIndices().concat(this.grid.GetBatchEditHelper().GetInsertedRecordKeys());
  for(var i = 0; i < indices.length; i++) {
   var visibleIndex = indices[i];
   rows = rows.concat(grid.GetBandedDataRows(visibleIndex));
   var extraRows = [ grid.GetGroupRow(visibleIndex), grid.GetPreviewRow(visibleIndex), grid.GetDetailRow(visibleIndex), grid.GetAdaptiveDetailRow(visibleIndex) ];
   rows = rows.concat(extraRows);
  }
  var pagerEmptyRows = ASPx.Data.CollectionToArray(grid.GetMainTable().getElementsByClassName(ASPx.GridViewConsts.EmptyPagerRowCssClass));
  rows = rows.concat(pagerEmptyRows);
  var levelCount = this.GetLayoutHeight();
  for(var i = 0; i <= levelCount; i++) {
   var groupFooterVisibleIndices = grid.GetGroupFooterVisibleIndices();
   for(var j = 0; j < groupFooterVisibleIndices.length; j++)
    rows.push(grid.GetGroupFooterRow(groupFooterVisibleIndices[j], i));
   rows.push(grid.GetFooterRow(i));
  }
  return rows;
 }
});
ASPx.GridViewLayoutHelper = GridViewLayoutHelper;
})();
